/* $Id$

This file is part of libmspdatafile
Copyright © 2006-2008, 2010  Mikko Rasa, Mikkosoft Productions
Distributed under the LGPL
*/

#include <msp/strings/formatter.h>
#include "loader.h"
#include "type.h"

using namespace std;

namespace {

bool signature_match(char s, char a)
{
	if(s==a)
		return true;
	if(s==Msp::DataFile::IntType::signature && a==Msp::DataFile::FloatType::signature)
		return true;
	return false;
}

bool signature_match(const string &st_sig, const string &act_sig)
{
	if(act_sig=="*")
		return true;
	else if(act_sig.size()==2 && act_sig[1]=='*')
	{
		for(string::const_iterator i=st_sig.begin(); i!=st_sig.end(); ++i)
			if(*i!=act_sig[0])
				return false;

		return true;
	}
	else if(st_sig.size()==act_sig.size())
	{
		for(unsigned i=0; i<st_sig.size(); ++i)
			if(!signature_match(st_sig[i], act_sig[i]))
				return false;

		return true;
	}
	else
		return false;
}

}

namespace Msp {
namespace DataFile {

Loader::~Loader()
{
	for(ActionMap::iterator i = actions.begin(); i!=actions.end(); ++i)
		delete i->second;
}

void Loader::load(Parser &p)
{
	while(p)
	{
		Statement st = p.parse();
		if(st.valid)
			load_statement(st);
	}
	finish();
}

void Loader::load(const Statement &st)
{
	for(list<Statement>::const_iterator i = st.sub.begin(); i!=st.sub.end(); ++i)
		load_statement(*i);
	finish();
}

void Loader::load_statement(const Statement &st)
{
	cur_st = &st;

	try
	{
		LoaderAction *act = find_action(ActionKey(st.keyword, st.get_signature()));
		if(act)
			act->execute(*this, st);
	}
	catch(Exception &e)
	{
		cur_st = 0;
		if(!e.where()[0])
			e.at(st.get_location());

		throw;
	}

	cur_st = 0;
}

void Loader::load_sub_with(Loader &ldr)
{
	if(!cur_st)
		throw InvalidState("load_sub called without current statement");

	ldr.load(*cur_st);
}

void Loader::add(const string &kwd, LoaderAction *act)
{
	ActionKey key(kwd, act->get_signature());
	ActionMap::iterator i = actions.find(key);
	if(i!=actions.end())
	{
		delete i->second;
		i->second = act;
	}
	else
		actions[key] = act;
}

LoaderAction *Loader::find_action(const ActionKey &key) const
{
	ActionMap::const_iterator begin = actions.lower_bound(ActionKey(key.keyword, string()));
	ActionMap::const_iterator end = actions.upper_bound(ActionKey(key.keyword, "~"));

	if(begin==end)
		throw KeyError("Unknown keyword", key.keyword);

	for(ActionMap::const_iterator i=begin; i!=end; ++i)
		if(signature_match(key.signature, i->first.signature))
			return i->second;

	throw KeyError(format("Keyword '%s' does not accept signature '%s'", key.keyword, key.signature));
}


Loader::ActionKey::ActionKey(const string &k, const string &s):
	keyword(k),
	signature(s)
{ }

bool Loader::ActionKey::operator<(const ActionKey &other) const
{
	if(keyword!=other.keyword)
		return keyword<other.keyword;
	return signature<other.signature;
}

} // namespace DataFile
} // namespace Msp
